import queue
import threading
import time
from typing import Any, Callable, Generator

from chia import instrumentation

END_MARKER = "THIS_IS_THE_END"


def _fancy_job_producer(
    job_queue: queue.Queue,
    job_generator: Callable[..., Generator[Any, None, None]],
    end_count: int,
    observable: instrumentation.Observable,
) -> None:
    observable.log_debug("Job producer starting")

    # Create generator instance
    job_generator_instance = job_generator()

    for job in job_generator_instance:
        job_queue.put(job)

    for _ in range(end_count):
        job_queue.put(END_MARKER)

    observable.log_debug("Job producer done")


def _fancy_job_consumer(
    id: int,
    job_queue: queue.Queue,
    result_queue: queue.Queue,
    task,
    observable,
):
    observable.log_debug(f"Job consumer #{id + 1} starting")
    while True:
        job = job_queue.get()
        if job == END_MARKER:
            result_queue.put(END_MARKER)
            break
        else:
            result = task(job)
            result_queue.put(result)

    observable.log_debug(f"Job consumer #{id + 1} done")


def _fancy_result_consumer(
    result_queue: queue.Queue, end_count: int, observable, timeout=1.0
):
    observable.log_debug("Result consumer starting")
    observed_end_markers = 0

    state = "OK"
    first_attempt_at = None
    waiting_time_reports = 0

    while True:
        try:
            # Attempt to get result
            result = result_queue.get(timeout=timeout)

            # See if we should stop
            if result == END_MARKER:
                observed_end_markers += 1
                if observed_end_markers == end_count:
                    break
                else:
                    continue

            # Don't stop, we have an item
            if state == "TIMEOUT_HIT":
                waiting_time = time.time() - first_attempt_at
                observable.log_info(
                    f"Threaded queue item finally retrieved. Waiting time: {waiting_time:.1f}s"
                )
                waiting_time_reports += 1
                if waiting_time_reports == 5:
                    observable.log_warning(
                        "Threaded queue hit timeout five times, will not report anymore"
                    )
                    state = "GIVEN_UP"
                else:
                    state = "OK"

            yield result

        except queue.Empty:
            if state == "OK":
                state = "TIMEOUT_HIT"
                observable.log_info("Threaded queue timeout hit")
                first_attempt_at = time.time() - timeout

    observable.log_debug("Result consumer done")


def threaded_processor(
    job_generator,
    task,
    observable,
    max_job_queue_size=4096,
    max_result_queue_size=64,
    num_threads=8,
):
    job_queue = queue.Queue(maxsize=max_job_queue_size)
    result_queue = queue.Queue(maxsize=max_result_queue_size)

    # Start job producer thread
    job_producer_thread = threading.Thread(
        target=_fancy_job_producer,
        args=(job_queue, job_generator, num_threads, observable),
    )

    job_producer_thread.start()

    # Start job consumer threads
    job_consumer_threads = []
    for i in range(num_threads):
        job_consumer_thread = threading.Thread(
            target=_fancy_job_consumer,
            args=(i, job_queue, result_queue, task, observable),
        )
        job_consumer_thread.start()
        job_consumer_threads += [job_producer_thread]

    # Consume and output results
    return _fancy_result_consumer(result_queue, num_threads, observable)



from chia.components.sample_transformers.sample_transformer import SampleTransformer


class IdentitySampleTransformer(SampleTransformer):
    def transform(self, samples, is_training: bool, label_resource_id: str):
        return samples



from chia import components
from chia.components.sample_transformers import identity
from chia.components.sample_transformers.sample_transformer import SampleTransformer


class SampleTransformerFactory(components.Factory):
    name_to_class_mapping = {"identity": identity.IdentitySampleTransformer}


__all__ = ["SampleTransformer", "SampleTransformerFactory"]



import json
import pathlib

import numpy as np
from PIL import Image

from chia import data, instrumentation, knowledge
from chia.components.datasets import dataset
from chia.helpers import NetworkResistantImage
from chia.knowledge import relation


class JSONDataset(dataset.Dataset, instrumentation.Observable):
    def __init__(self, json_path, base_path, side_length, use_lazy_mode=True):
        instrumentation.Observable.__init__(self)

        self._json_path = json_path
        self._base_path = base_path
        self._side_length = side_length
        self._use_lazy_mode = use_lazy_mode

        # Read JSON file
        with open(self._json_path) as json_file:
            self._json_data = json.load(json_file)

        # Check format
        if self._json_data["file_format"] != "JSONDataset.v1":
            raise ValueError("Unsupported JSON dataset file format!")

        # Knowledge
        self._prediction_targets = set(self._json_data["prediction_targets"])
        self._hyponymy_relation_data = self._json_data["hyponymy_relation"]
        self._namespace = self._json_data["namespace"]

        # Samples
        self._train_pools = self._json_data["train_pools"]
        self._test_pools = self._json_data["test_pools"]

    def setup(self, **kwargs):
        pass

    def train_pool_count(self):
        return len(self._test_pools)

    def test_pool_count(self):
        return len(self._train_pools)

    def train_pool(self, index, label_resource_id):
        return [
            self._build_sample(sample_dict, label_resource_id)
            for sample_dict in self._train_pools[index]
        ]

    def test_pool(self, index, label_resource_id):
        return [
            self._build_sample(sample_dict, label_resource_id)
            for sample_dict in self._test_pools[index]
        ]

    def namespace(self):
        return self._namespace

    def get_hyponymy_relation_source(self) -> knowledge.RelationSource:
        return relation.StaticRelationSource(self._hyponymy_relation_data)

    def prediction_targets(self):
        return self._prediction_targets

    def _build_sample(self, sample_dict: dict, label_resource_id: str) -> data.Sample:
        image_location = pathlib.Path(self._base_path) / sample_dict["image_location"]
        uid = sample_dict["uid"]
        label_gt = sample_dict["label_gt"]
        sample = data.Sample(source=self.__class__.__name__, uid=uid)
        sample = sample.add_resource(
            self.__class__.__name__, label_resource_id, label_gt
        )

        if self._use_lazy_mode:
            sample = sample.add_resource(
                self.__class__.__name__,
                "image_location",
                image_location,
            ).add_lazy_resource(
                self.__class__.__name__, "input_img_np", self._load_from_location
            )
        else:
            sample = sample.add_resource(
                self.__class__.__name__,
                "image_location",
                image_location,
            )
            sample = sample.add_resource(
                self.__class__.__name__,
                "input_img_np",
                self._load_from_location(sample),
            )

        return sample

    def _load_from_location(self, sample_):
        im = NetworkResistantImage.open(
            sample_.get_resource("image_location"), self
        ).resize((self._side_length, self._side_length), Image.ANTIALIAS)
        if im.mode != "RGB":
            im = im.convert("RGB")
        return np.asarray(im)
